"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
var _pouchdbAdapterUtils = require("pouchdb-adapter-utils");
var _pouchdbMerge = require("pouchdb-merge");
var _pouchdbJson = require("pouchdb-json");
var _pouchdbErrors = require("pouchdb-errors");
var _constants = require("./constants.js");
var _utils = require("./utils.js");
var _debug = require("./debug.js");
async function sqliteBulkDocs(dbOpts, req, opts, api, transaction, sqliteChanges) {
  const newEdits = opts.new_edits;
  const userDocs = req.docs;
  const docInfos = userDocs.map(doc => {
    if (doc._id && (0, _pouchdbAdapterUtils.isLocalId)(doc._id)) {
      return doc;
    }
    return (0, _pouchdbAdapterUtils.parseDoc)(doc, newEdits, dbOpts);
  });
  const docInfoErrors = docInfos.filter(docInfo => docInfo.error);
  if (docInfoErrors.length) {
    throw docInfoErrors[0];
  }
  let tx;
  const results = new Array(docInfos.length);
  const fetchedDocs = new Map();
  async function verifyAttachment(digest) {
    _debug.logger.debug('verify attachment:', digest);
    const sql = 'SELECT count(*) as cnt FROM ' + _constants.ATTACH_STORE + ' WHERE digest=?';
    const result = await tx.execute(sql, [digest]);
    if (result.rows[0]?.cnt === 0) {
      const err = (0, _pouchdbErrors.createError)(_pouchdbErrors.MISSING_STUB, 'unknown stub attachment with digest ' + digest);
      _debug.logger.error('unknown:', err);
      throw err;
    } else {
      _debug.logger.debug('ok');
      return true;
    }
  }
  async function verifyAttachments() {
    const digests = [];
    docInfos.forEach(docInfo => {
      if (docInfo.data && docInfo.data._attachments) {
        Object.keys(docInfo.data._attachments).forEach(filename => {
          const att = docInfo.data._attachments[filename];
          if (att.stub) {
            _debug.logger.debug('attachment digest', att.digest);
            digests.push(att.digest);
          }
        });
      }
    });
    if (!digests.length) return;
    for (const digest of digests) {
      await verifyAttachment(digest);
    }
  }
  async function writeDoc(docInfo, winningRev, _winningRevIsDeleted, newRevIsDeleted, isUpdate, _delta, resultsIdx) {
    _debug.logger.debug('writeDoc:', {
      ...docInfo,
      data: null
    });
    async function dataWritten(tx, seq) {
      const id = docInfo.metadata.id;
      let revsToCompact = docInfo.stemmedRevs || [];
      if (isUpdate && api.auto_compaction) {
        revsToCompact = (0, _pouchdbMerge.compactTree)(docInfo.metadata).concat(revsToCompact);
      }
      if (revsToCompact.length) {
        (0, _utils.compactRevs)(revsToCompact, id, tx);
      }
      docInfo.metadata.seq = seq;
      const rev = docInfo.metadata.rev;
      delete docInfo.metadata.rev;
      const sql = isUpdate ? 'UPDATE ' + _constants.DOC_STORE + ' SET json=?, max_seq=?, winningseq=' + '(SELECT seq FROM ' + _constants.BY_SEQ_STORE + ' WHERE doc_id=' + _constants.DOC_STORE + '.id AND rev=?) WHERE id=?' : 'INSERT INTO ' + _constants.DOC_STORE + ' (id, winningseq, max_seq, json) VALUES (?,?,?,?);';
      const metadataStr = (0, _pouchdbJson.safeJsonStringify)(docInfo.metadata);
      const params = isUpdate ? [metadataStr, seq, winningRev, id] : [id, seq, seq, metadataStr];
      await tx.execute(sql, params);
      results[resultsIdx] = {
        ok: true,
        id: docInfo.metadata.id,
        rev: rev
      };
      fetchedDocs.set(id, docInfo.metadata);
    }
    async function insertAttachmentMappings(seq) {
      const attsToAdd = Object.keys(data._attachments || {});
      if (!attsToAdd.length) {
        return;
      }
      function add(att) {
        const sql = 'INSERT INTO ' + _constants.ATTACH_AND_SEQ_STORE + ' (digest, seq) VALUES (?,?)';
        const sqlArgs = [data._attachments[att].digest, seq];
        return tx.execute(sql, sqlArgs);
      }
      await Promise.all(attsToAdd.map(att => add(att)));
    }
    docInfo.data._id = docInfo.metadata.id;
    docInfo.data._rev = docInfo.metadata.rev;
    const attachments = Object.keys(docInfo.data._attachments || {});
    if (newRevIsDeleted) {
      docInfo.data._deleted = true;
    }
    for (const key of attachments) {
      const att = docInfo.data._attachments[key];
      if (!att.stub) {
        const data = att.data;
        delete att.data;
        att.revpos = parseInt(winningRev, 10);
        const digest = att.digest;
        await saveAttachment(digest, data);
      }
    }
    const data = docInfo.data;
    const deletedInt = newRevIsDeleted ? 1 : 0;
    const id = data._id;
    const rev = data._rev;
    const json = (0, _utils.stringifyDoc)(data);
    const sql = 'INSERT INTO ' + _constants.BY_SEQ_STORE + ' (doc_id, rev, json, deleted) VALUES (?, ?, ?, ?);';
    const sqlArgs = [id, rev, json, deletedInt];
    try {
      const result = await tx.execute(sql, sqlArgs);
      const seq = result.insertId;
      if (typeof seq === 'number') {
        await insertAttachmentMappings(seq);
        await dataWritten(tx, seq);
      }
    } catch (e) {
      // constraint error, recover by updating instead (see #1638)
      // https://github.com/pouchdb/pouchdb/issues/1638
      const fetchSql = (0, _utils.select)('seq', _constants.BY_SEQ_STORE, null, 'doc_id=? AND rev=?');
      const res = await tx.execute(fetchSql, [id, rev]);
      const seq = res.rows[0].seq;
      _debug.logger.debug(`Got a constraint error, updating instead: seq=${seq}, id=${id}, rev=${rev}`);
      const sql = 'UPDATE ' + _constants.BY_SEQ_STORE + ' SET json=?, deleted=? WHERE doc_id=? AND rev=?;';
      const sqlArgs = [json, deletedInt, id, rev];
      await tx.execute(sql, sqlArgs);
      await insertAttachmentMappings(seq);
      await dataWritten(tx, seq);
    }
  }
  function websqlProcessDocs() {
    return new Promise((resolve, reject) => {
      let chain = Promise.resolve();
      (0, _pouchdbAdapterUtils.processDocs)(dbOpts.revs_limit, docInfos, api, fetchedDocs, tx, results, (docInfo, winningRev, winningRevIsDeleted, newRevIsDeleted, isUpdate, delta, resultsIdx, callback) => {
        chain = chain.then(() => {
          return writeDoc(docInfo, winningRev, winningRevIsDeleted, newRevIsDeleted, isUpdate, delta, resultsIdx).then(callback, callback);
        });
      }, opts, err => {
        if (!err) resolve();else reject(err);
      });
    });
  }
  async function fetchExistingDocs() {
    if (!docInfos.length) return;
    for (const docInfo of docInfos) {
      if (docInfo._id && (0, _pouchdbAdapterUtils.isLocalId)(docInfo._id)) {
        continue;
      }
      const id = docInfo.metadata.id;
      const result = await tx.execute('SELECT json FROM ' + _constants.DOC_STORE + ' WHERE id = ?', [id]);
      if (result.rows?.length) {
        const metadata = (0, _pouchdbJson.safeJsonParse)(result.rows[0].json);
        fetchedDocs.set(id, metadata);
      }
    }
  }
  async function saveAttachment(digest, data) {
    _debug.logger.debug('saveAttachment:', digest);
    let sql = 'SELECT digest FROM ' + _constants.ATTACH_STORE + ' WHERE digest=?';
    const result = await tx.execute(sql, [digest]);
    if (result.rows?.length) return;
    sql = 'INSERT INTO ' + _constants.ATTACH_STORE + ' (digest, body, escaped) VALUES (?,?,0)';
    await tx.execute(sql, [digest, data]);
  }
  await new Promise((resolve, reject) => {
    (0, _pouchdbAdapterUtils.preprocessAttachments)(docInfos, 'binary', err => {
      if (err) reject(err);else resolve();
    });
  });
  await transaction(async txn => {
    await verifyAttachments();
    try {
      tx = txn;
      await fetchExistingDocs();
      await websqlProcessDocs();
      sqliteChanges.notify(api._name);
    } catch (err) {
      throw (0, _utils.handleSQLiteError)(err);
    }
  });
  return results;
}
var _default = exports.default = sqliteBulkDocs;
//# sourceMappingURL=bulkDocs.js.map